"""
Proxy server for Libresprite.
"""

import threading
from typing import Optional
from mcp.server.fastmcp import Context
from flask import Flask, jsonify, request

class LibrespriteProxy:
    """
    Proxy server for Libresprite.

    This is a relay server that acts as a bridge between the MCP server and the remote libresprite script.
    From the POV of the MCP server, this fully abstracts libresprite into a convenient IO interface to execute remote scripts.
    """

    def __init__(self, port: int):
        # Stores current script
        self._script: str | None = None

        # Stores output of the script execution
        self._output: str | None = None

        # Stores flag indicating if execution is in progress
        self._lock: bool = False

        # Initialize event handlers
        self._script_event = threading.Event()
        self._output_event = threading.Event()

        # Relay server configuration
        self.port = port
        self.app = Flask(__name__)

        # Initialize server routes
        self._setup_routes()
        self._server_thread: Optional[threading.Thread] = None

    def _setup_routes(self):
        """Setup Flask routes."""

        @self.app.get('/')
        def get_script():
            """Get the current script."""
            if self._script is None:
                self._script_event.wait()
                self._script_event.clear()
            script = self._script
            self._script = None
            return jsonify({"script": script})

        @self.app.post('/')
        def post_output():
            """Post execution output."""
            if not self._lock:
                # ignore random requests
                return jsonify({"status": "ignored"})
            req = request.get_json(force=True, silent=True)
            if req:
                output = req.get('output')
            else:
                return jsonify({"status": "invalid"})
            self._output = output
            self._output_event.set()
            return jsonify({"status": "success"})

        @self.app.get('/ping')
        def ping():
            """Ping endpoint for health checks."""
            return jsonify({"status": "pong"})

    def _run_server(self):
        """Run the Flask server."""
        self.app.run(
            host='localhost',
            port=self.port,
            debug=False,
            use_reloader=False
        )

    def start(self):
        """Start the HTTP server in a background thread."""
        if self._server_thread and self._server_thread.is_alive():
            return

        self._server_thread = threading.Thread(
            target=self._run_server,
            daemon=True
        )
        self._server_thread.start()

    def run_script(self, script: str, ctx: Context) -> str:
        """
        Run a script in the execution context.

        WARNING: This method is synchronous and blocking.

        Args:
            script: The script to execute
        """
        # This proxy only allows one script to be executed at a time
        if self._lock:
            ctx.error("Script execution is already in progress...")
            raise RuntimeError("Script execution is already in progress.")

        # Sending the script
        ctx.info("Sending script to libresprite...")
        self._lock = True
        self._script = script
        self._script_event.set()

        # Waiting for execution
        if not self._output_event.wait(timeout=15):
            ctx.warning("This is taking longer than usual, make sure the user has the Libresprite application with the remote script running?")
        self._output_event.wait()
        self._output_event.clear()

        # Return the output
        ctx.info("Script execution completed, checking the logs...")
        output = self._output
        self._output = None
        self._lock = False
        return output


"""
MCP server implementation that exposes tools for interacting with the libresprite-proxy server.
"""

import os
from mcp.server.fastmcp import FastMCP, Context
from .libresprite_proxy import LibrespriteProxy


class MCPServer:
    """The LibreSprite MCP Server."""

    def __init__(self, libresprite_proxy: LibrespriteProxy, server_name: str = "libresprite"):
        # Cache the libresprite proxy instance
        self._libresprite_proxy = libresprite_proxy

        # Initialize FastMCP
        self.mcp = FastMCP(server_name)

        # Setup MCP tools, prompts, and resources
        self._setup_tools()
        self._setup_resources()
        self._setup_prompts()

    def _setup_tools(self):
        """Setup MCP tools."""

        @self.mcp.tool()
        def run_script(script: str, ctx: Context) -> str:
            """
            Run a JavaScript script inside Libresprite.

            IMPORTANT: Make sure you are well versed with the documentation and examples provided in the resources `docs:reference` and `docs:examples`.

            Args:
                script: The script to execute

            Returns:
                Console output
            """
            return self._libresprite_proxy.run_script(script, ctx)

    def _setup_resources(self):
        """Setup MCP resources."""

        base_dir = os.path.dirname(os.path.abspath(__file__))

        @self.mcp.resource("docs://reference")
        def read_reference() -> str:
            """Read the libresprite command reference documentation."""
            doc_path = os.path.join(base_dir, "resources", "reference.txt")
            try:
                with open(doc_path, "r", encoding="utf-8") as f:
                    return f.read()
            except Exception as e:
                return f"Error reading reference.txt: {e}"

        @self.mcp.resource("docs://examples")
        def read_examples() -> str:
            """Read example scripts using libresprite commands."""
            example_path = os.path.join(base_dir, "resources", "examples.txt")
            try:
                with open(example_path, "r", encoding="utf-8") as f:
                    return f.read()
            except Exception as e:
                return f"Error reading examples.txt: {e}"

    def _setup_prompts(self):
        """Setup MCP prompts."""

        @self.mcp.prompt(title="libresprite")
        def libresprite(prompt: str) -> str:
            """
            Prompt template to use the libresprite tool with proper context conditioning.

            Args:
                prompt: User prompt

            Returns:
                Prompt to process
            """
            return f"""
            Libresprite is a program for creating and editing pixel art and animations using JavaScript.

            Before proceeding, please ensure you are well versed with the documentation and examples provided in the resources `docs:reference` and `docs:examples`.

            You can use the `run_script` tool to execute JavaScript scripts in the context of libresprite.

            Here's what you need to do using the above tools and resources:

            {prompt}
            """

    def run(self, transport: str = 'stdio'):
        """Run the MCP server."""
        self.mcp.run(transport=transport)


"""
Main entry point for the Libresprite MCP server application.
"""

import click
import logging
from .libresprite_proxy import LibrespriteProxy
from .mcp_server import MCPServer

def main() -> None:
    """Main entry point for the application."""
    # Disable unwanted logging to avoid messing with stdio
    logging.disable(logging.WARN)
    # HACK: https://stackoverflow.com/a/57086684
    def secho(text, file=None, nl=None, err=None, color=None, **styles):
        pass
    def echo(text, file=None, nl=None, err=None, color=None, **styles):
        pass
    click.echo = echo
    click.secho = secho

    # Initialize and start HTTP relay server
    libresprite_proxy = LibrespriteProxy(port=64823)
    libresprite_proxy.start()

    # Initialize and run MCP server (this blocks)
    mcp_server = MCPServer(libresprite_proxy)
    mcp_server.run(transport='stdio')

if __name__ == '__main__':
    main()